from typing import (
    Dict,
    List,
    Optional,
    TypedDict,
    Union,
)
from enum import Enum
from webviz_config.utils import callback_typecheck, ConversionError


def test_callback_typecheck() -> None:
    class MyEnum(str, Enum):
        VALUE_1 = "value-1"

    class MyTypedDict(TypedDict):
        name: str
        year: int

    class DeepTypedDict(TypedDict):
        value: MyTypedDict

    ############################################################

    def expect_none(arg: None) -> None:
        assert arg is None

    callback_typecheck(expect_none)(None)

    ############################################################

    def expect_enum(arg: MyEnum) -> None:
        assert isinstance(arg, MyEnum)

    callback_typecheck(expect_enum)("value-1")

    ############################################################

    def expect_typed_dict(arg: MyTypedDict) -> None:
        types = [type(value) for value in arg.values()]
        assert (
            isinstance(arg, dict)
            and set(arg.keys())
            == set(MyTypedDict.__annotations__.keys())  # pylint: disable=no-member
            and set(types)
            == set(MyTypedDict.__annotations__.values())  # pylint: disable=no-member
        )

    callback_typecheck(expect_typed_dict)({"name": "Name", "year": 1990})

    # If any invalid key is given to a `TypedDict`, assert that a `ConversionError` is raised
    try:
        callback_typecheck(expect_typed_dict)({"name": "Name", "year2": 1990})
        assert False
    except ConversionError:
        pass
    except Exception:  # pylint: disable=broad-except
        assert False

    ############################################################

    def expect_deep_typed_dict(arg: DeepTypedDict) -> None:
        types = [type(value) for value in arg.values()]
        assert (
            isinstance(arg, dict)
            and set(arg.keys())
            == set(DeepTypedDict.__annotations__.keys())  # pylint: disable=no-member
            # NOTE: A `TypedDict` is a `dict` at runtime
            and set(types) == set([dict])
            and set(arg["value"].keys())
            == set(MyTypedDict.__annotations__.keys())  # pylint: disable=no-member
        )

    callback_typecheck(expect_deep_typed_dict)(
        {"value": {"name": "Name", "year": 1990}}
    )

    ############################################################

    def expect_int(arg: int) -> None:
        assert isinstance(arg, int)

    callback_typecheck(expect_int)("1")

    ############################################################

    def expect_float(arg: float) -> None:
        assert isinstance(arg, float)

    callback_typecheck(expect_float)("1.5")

    ############################################################

    def expect_str(arg: str) -> None:
        assert isinstance(arg, str)

    callback_typecheck(expect_str)(1)

    ############################################################

    def expect_list(arg: List[str]) -> None:
        assert isinstance(arg, list)
        assert isinstance(arg[0], str)

    callback_typecheck(expect_list)([1, 2, 3])

    ############################################################

    def expect_dict(arg: Dict[str, int]) -> None:
        assert isinstance(arg, dict)
        assert isinstance(list(arg.values())[0], int)
        assert isinstance(list(arg.keys())[0], str)

    callback_typecheck(expect_dict)({"1": 1})

    ############################################################

    def expect_optional(arg: Optional[str]) -> Optional[str]:
        return arg

    assert callback_typecheck(expect_optional)(None) is None
    assert isinstance(callback_typecheck(expect_optional)("string"), str)

    ############################################################

    def expect_union(arg: Union[str, int]) -> Union[str, int]:
        return arg

    assert isinstance(callback_typecheck(expect_union)("1"), str)
    assert isinstance(callback_typecheck(expect_union)(1), int)
    assert isinstance(callback_typecheck(expect_union)(1.5), str)

    ############################################################



import os
import re
from typing import List, Callable, Optional

from webviz_config.utils import terminal_colors


def terminal_title(title: str) -> None:
    print("\n" + terminal_colors.BOLD + title + terminal_colors.END)
    print("=" * len(title) + "\n")


def terminal_yes_no(question: str) -> bool:
    while True:
        reply = input(question + " (y/n): ").lower().strip()
        if reply.startswith("y"):
            return True
        if reply.startswith("n"):
            return False


def terminal_list_selector(options: List[str], display_string: str) -> str:

    for i, subscription in enumerate(options):
        print(f"{i+1}) {subscription}")
    print()

    selected_value = None
    while selected_value not in options:
        try:
            index = int(input(f"Choose {display_string} index to use: "))
            return_value = selected_value = options[index - 1]
        except (IndexError, ValueError):
            print(f"Select an integer in range 1-{len(options)}.")

    print(
        "\n"
        + terminal_colors.BOLD
        + f"Selected {display_string}: {selected_value}"
        + terminal_colors.END
    )
    return return_value


def user_input_from_stdin(env_var: str, noun: str, regex: str = None) -> str:
    if env_var in os.environ:
        value = os.environ[env_var]
        print(f"Using {noun} '{value}' (from environment variable {env_var}).")
        if regex is not None and not re.fullmatch(regex, value):
            raise RuntimeError("Input not on expected format.")
        return value

    print(f"Tip: Default value can be set through env. variable {env_var}\n")
    while True:
        value = input(f"{noun.capitalize()}: ").strip()
        if regex is None or re.fullmatch(regex, value):
            return value
        print("Input not on expected format.")


def user_input_from_list(env_var: str, noun: str, choices: List[str]) -> str:
    if env_var in os.environ:
        value = os.environ[env_var]
        print(f"Using {noun} '{value}' (from environment variable {env_var}).")
        if value not in choices:
            raise ValueError(
                f"Value '{value}' not among available choices ({', '.join(choices)})"
            )
        return value

    print(f"Tip: Default value can be set through env. variable {env_var}\n")
    return terminal_list_selector(choices, noun)


def user_input_optional_reuse(
    env_var: str,
    noun: str,
    exists_function: Callable,
    reuse_allowed: bool = True,
    regex: Optional[str] = None,
) -> str:
    if env_var in os.environ:
        value = os.environ[env_var]
        print(f"Using {noun} '{value}' (from environment variable {env_var}).")
        if not reuse_allowed and exists_function(value):
            raise RuntimeError(f"There already exists an {noun} with name '{value}'.")
        if regex is not None and not re.match(regex, value):
            raise RuntimeError(
                f"Value {value} does not satisfy regular expression {regex}"
            )
        return value

    print(f"Tip: Default value can be set through env. variable {env_var}\n")

    while True:
        value = input(f"{noun.capitalize()}: ")

        if regex is not None and not re.match(regex, value):
            print(f"Value {value} does not satisfy regular expression {regex}")
            continue

        check = exists_function(value)

        if isinstance(check, tuple):
            exists, message = check
        else:
            exists = check
            message = None

        if exists:
            if not reuse_allowed:
                print(
                    f"There already exists an {noun} with name '{value}'"
                    if message is None
                    else message
                )
                continue
            print(f"Found existing {noun} with value '{value}'")
            reuse = terminal_yes_no(f"Do you want to reuse the existing {noun}")
            if reuse:
                return value
            continue

        print(f"{noun.capitalize()} name '{value}' is available.")
        return value



from unittest import mock

import pandas as pd
import dash

from webviz_config.common_cache import CACHE
from webviz_config.generic_plugins import _data_table

# mocked functions
GET_DATA = "webviz_config.generic_plugins._data_table.get_data"


def test_data_table(dash_duo):

    app = dash.Dash(__name__)
    app.config.suppress_callback_exceptions = True
    CACHE.init_app(app.server)
    code_file = "./tests/data/example_data.csv"
    with mock.patch(GET_DATA) as mock_path:
        mock_path.return_value = pd.read_csv(code_file)
        page = _data_table.DataTable(code_file)
        app.layout = page.layout
        dash_duo.start_server(app)
        assert dash_duo.get_logs() == [], "browser console should contain no error"


def test_data_table_with_settings(dash_duo):

    app = dash.Dash(__name__)
    app.css.config.serve_locally = True
    app.scripts.config.serve_locally = True
    app.config.suppress_callback_exceptions = True
    CACHE.init_app(app.server)
    code_file = "./tests/data/example_data.csv"
    with mock.patch(GET_DATA) as mock_path:
        mock_path.return_value = pd.read_csv(code_file)
        page = _data_table.DataTable(
            csv_file=code_file, sorting=False, filtering=False, pagination=False
        )
        app.layout = page.layout
        dash_duo.start_server(app)
        assert dash_duo.get_logs() == [], "browser console should contain no error"



from typing import Any, Dict, List, Optional, Tuple, Union, Type
from dataclasses import dataclass
import sys

from dash.development.base_component import Component
from dash import html, Input, Output, State, callback
from dash.exceptions import PreventUpdate

import pandas as pd

import webviz_core_components as wcc

from webviz_config import WebvizPluginABC, EncodedFile

from webviz_config.webviz_plugin_subclasses import (
    ViewABC,
    ViewElementABC,
    SettingsGroupABC,
)
from webviz_config.utils import callback_typecheck, StrEnum

from webviz_config.generic_plugins._example_wlf_plugin._shared_view_elements import (
    TextViewElement,
)

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated


class Kindness(StrEnum):
    FRIENDLY = "friendly"
    UNFRIENDLY = "unfriendly"


class Coordinates(StrEnum):
    XY = "xy"
    YX = "yx"


class PlotViewSettingsGroup(SettingsGroupABC):
    class Ids(StrEnum):
        COORDINATES_SELECTOR = "coordinates-selector"

    def __init__(self) -> None:
        super().__init__("Plot coordinate system")

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                id=self.register_component_unique_id(
                    PlotViewSettingsGroup.Ids.COORDINATES_SELECTOR
                ),
                label="Coordinates",
                options=[
                    {
                        "label": "x - y",
                        "value": Coordinates.XY,
                    },
                    {
                        "label": "y - x",
                        "value": Coordinates.YX,
                    },
                ],
                value=Coordinates.XY,
                clearable=False,
            )
        ]


class PlotViewElement(ViewElementABC):
    class Ids(StrEnum):
        GRAPH = "graph"

    def __init__(self, data: List[Tuple[int, int]]) -> None:
        super().__init__(flex_grow=8)
        self.data = data

    def inner_layout(self) -> Union[str, Type[Component]]:
        return html.Div(
            style={"height": "20vh"},
            children=[
                wcc.Graph(
                    id=self.register_component_unique_id(PlotViewElement.Ids.GRAPH),
                    figure={
                        "data": [
                            {
                                "x": [x[0] for x in self.data],
                                "y": [x[1] for x in self.data],
                            }
                        ],
                        "layout": {
                            "title": "Example Graph Swapped",
                        },
                    },
                    config={
                        "responsive": True,
                    },
                )
            ],
        )

    @staticmethod
    def download_data_df(graph_figure: Dict[str, Any]) -> pd.DataFrame:
        graph_data: Optional[List[Dict[str, Any]]] = graph_figure.get("data", None)
        if not graph_data:
            return "No data present in graph figure"

        x_values = graph_data[0].get("x", None)
        y_values = graph_data[0].get("y", None)
        if x_values is None or y_values is None:
            return f"Missing x or y data: x = {x_values} and y = {y_values}"

        df = pd.DataFrame(
            columns=["x", "y"],
        )
        df["x"] = x_values
        df["y"] = y_values
        return df

    def compressed_plugin_data(
        self, graph_figure: Dict[str, Any]
    ) -> Union[EncodedFile, str]:
        return WebvizPluginABC.plugin_data_compress(
            [
                {
                    "filename": f"""{
                        self.component_unique_id(PlotViewElement.Ids.GRAPH).to_string()
                    }.csv""",
                    "content": self.download_data_df(graph_figure).to_csv(index=False),
                }
            ]
        )

    def set_callbacks(self) -> None:
        @callback(
            self.view_element_data_output(),
            self.view_element_data_requested(),
            State(
                self.component_unique_id(PlotViewElement.Ids.GRAPH).to_string(),
                "figure",
            ),
        )
        def _download_data(
            data_requested: Union[int, None], graph_figure: Dict[str, Any]
        ) -> Union[EncodedFile, str]:
            if data_requested is None:
                raise PreventUpdate

            return self.compressed_plugin_data(graph_figure)


class PlotView(ViewABC):
    class Ids(StrEnum):
        TEXT = "text"
        PLOT = "plot"
        PLOT_SETTINGS = "plot-settings"

    @dataclass
    class Slots:
        kindness_selector: Annotated[Input, str]
        power_selector: Annotated[Input, str]

    def __init__(self, data: List[Tuple[int, int]], slots: Slots) -> None:
        super().__init__("Plot")
        self.data = data

        self.slots = slots

        self._plot_view = PlotViewElement(self.data)

        row = self.add_row()
        row.add_view_element(TextViewElement(), PlotView.Ids.TEXT)
        row.add_view_element(self._plot_view, PlotView.Ids.PLOT)

        self.add_settings_group(PlotViewSettingsGroup(), PlotView.Ids.PLOT_SETTINGS)

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.view_element(PlotView.Ids.TEXT)
                .component_unique_id(TextViewElement.Ids.TEXT)
                .to_string(),
                "children",
            ),
            self.slots.kindness_selector,
        )
        @callback_typecheck
        def _change_kindness(kindness: Kindness) -> Component:
            if kindness == Kindness.FRIENDLY:
                return [
                    html.H1("Hello"),
                    """
                    I am an example plugin.
                    Please have a look how views and settings are working in my environment =).
                    """,
                ]

            return [
                html.H1("Goodbye"),
                "I am a bloody example plugin. Leave me alone! =(",
            ]

        @callback(
            Output(
                self.view_elements()[1]
                .component_unique_id(PlotViewElement.Ids.GRAPH)
                .to_string(),
                "figure",
            ),
            self.slots.power_selector,
            Input(
                self.settings_group(PlotView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotViewSettingsGroup.Ids.COORDINATES_SELECTOR)
                .to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _change_power_and_coordinates(power: int, coordinates: Coordinates) -> dict:
            layout = {
                "title": "Example Graph Swapped",
                "margin": {"t": 50, "r": 20, "b": 20, "l": 20},
            }
            if power == 2:
                self.data = [(x, x * x) for x in range(0, 10)]
            else:
                self.data = [(x, x * x * x) for x in range(0, 10)]

            if coordinates == Coordinates.YX:
                return {
                    "data": [
                        {
                            "x": [x[1] for x in self.data],
                            "y": [x[0] for x in self.data],
                        }
                    ],
                    "layout": layout,
                }
            return {
                "data": [
                    {
                        "x": [x[0] for x in self.data],
                        "y": [x[1] for x in self.data],
                    }
                ],
                "layout": layout,
            }

        @callback(
            self.view_data_output(),
            self.view_data_requested(),
            State(
                self._plot_view.component_unique_id(
                    PlotViewElement.Ids.GRAPH
                ).to_string(),
                "figure",
            ),
        )
        def _download_data(
            data_requested: Union[int, None],
            graph_figure: Dict[str, Any],
        ) -> Union[EncodedFile, str]:
            if data_requested is None:
                raise PreventUpdate

            return self._plot_view.compressed_plugin_data(graph_figure)



# pylint: disable=line-too-long
from typing import Any, Callable, get_origin, _TypedDictMeta, TypeVar, Union  # type: ignore[attr-defined]
import inspect

T = TypeVar("T")


class ConversionError(Exception):
    pass


def convert(arg: Any, convert_to: T) -> T:
    # pylint: disable=too-many-return-statements, too-many-branches
    additional_error_message: str = ""
    try:
        if convert_to is None and arg is None:
            return None
        if inspect.isclass(convert_to) and not isinstance(convert_to, _TypedDictMeta):
            return convert_to(arg)
        if (
            isinstance(convert_to, _TypedDictMeta)
            and "__annotations__" in dir(convert_to)
            and isinstance(arg, dict)
        ):
            new_dict = convert_to()
            for key, value in arg.items():
                if key in list(convert_to.__annotations__.keys()):
                    new_dict[key] = convert(value, convert_to.__annotations__[key])
                else:
                    raise Exception(
                        f"""
                        Key '{key}' not allowed in '{convert_to}'.\n
                        Allowed keys are: {', '.join(list(convert_to.__annotations__.keys()))}
                        """
                    )

            if not convert_to.__total__ or len(new_dict.keys()) == len(
                convert_to.__annotations__.keys()
            ):
                return new_dict

        if convert_to is list and isinstance(arg, list):
            return arg  # type: ignore[return-value]
        if get_origin(convert_to) is list and isinstance(arg, list):
            if "__args__" in dir(convert_to):
                return [convert(a, convert_to.__args__[0]) for a in arg]  # type: ignore[attr-defined,return-value]
        if convert_to is dict and isinstance(arg, dict):
            return arg  # type: ignore[return-value]
        if get_origin(convert_to) is dict and isinstance(arg, dict):
            if "__args__" in dir(convert_to) and len(convert_to.__args__) == 2:  # type: ignore[attr-defined]
                return {  # type: ignore[return-value]
                    convert(key, convert_to.__args__[0]): convert(  # type: ignore[attr-defined]
                        value, convert_to.__args__[1]  # type: ignore[attr-defined]
                    )
                    for key, value in arg.items()
                }
        if get_origin(convert_to) is Union:
            if "__args__" in dir(convert_to):
                for convert_type in convert_to.__args__:  # type: ignore[attr-defined]
                    if isinstance(arg, convert_type):
                        return arg
                for convert_type in convert_to.__args__:  # type: ignore[attr-defined]
                    try:
                        return convert(arg, convert_type)
                    except ConversionError:
                        pass

    # pylint: disable=broad-except
    except Exception as exception:
        additional_error_message = f"\n\n{exception}"

    raise ConversionError(
        f"Argument of type '{type(arg)}' cannot be converted to type '{convert_to}'.{additional_error_message}"
    )


def callback_typecheck(func: Callable) -> Callable:
    signature = inspect.signature(func)
    argument_annotations: list = []

    for param in signature.parameters.values():
        argument_annotations.append(param.annotation)

    def wrapper(*_args) -> signature.return_annotation:  # type: ignore[no-untyped-def,name-defined]
        adjusted_args: list = []

        for index, arg in enumerate(_args):
            adjusted_args.append(convert(arg, argument_annotations[index]))

        return func(*adjusted_args)

    return wrapper



from dataclasses import dataclass
from typing import Dict, List, Tuple, Type, Union
import sys

from dash.development.base_component import Component
from dash import html, Input, Output, State, dash_table, callback
from dash.exceptions import PreventUpdate

import pandas as pd

import webviz_core_components as wcc

from webviz_config.utils import callback_typecheck, StrEnum
from webviz_config import WebvizPluginABC, EncodedFile

from webviz_config.webviz_plugin_subclasses import (
    ViewABC,
    ViewElementABC,
    SettingsGroupABC,
)

from webviz_config.generic_plugins._example_wlf_plugin._shared_view_elements import (
    TextViewElement,
)

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated


class Kindness(StrEnum):
    FRIENDLY = "friendly"
    UNFRIENDLY = "unfriendly"


class Order(StrEnum):
    ASC = "asc"
    DESC = "desc"


class TableViewElement(ViewElementABC):
    class Ids(StrEnum):
        TABLE = "table"

    def __init__(self, data: List[Tuple[int, int]]) -> None:
        super().__init__()
        self.data = data

    def inner_layout(self) -> Union[str, Type[Component]]:
        return dash_table.DataTable(
            id=self.register_component_unique_id(TableViewElement.Ids.TABLE),
            columns=[{"id": "x", "name": "X"}, {"id": "y", "name": "Y"}],
            data=[{"x": d[0], "y": d[1]} for d in self.data],
        )

    @staticmethod
    def download_data_df(table_data: List[Dict[str, int]]) -> pd.DataFrame:
        x_values = []
        y_values = []
        for index, elm in enumerate(table_data):
            _x = elm.get("x", None)
            _y = elm.get("y", None)
            if _x is None:
                raise ValueError(f'No "x" column data in table at index {index}')
            if _y is None:
                raise ValueError(f'No "y" column data in table at index {index}')
            x_values.append(_x)
            y_values.append(_y)

        df = pd.DataFrame(
            columns=["x", "y"],
        )
        df["x"] = x_values
        df["y"] = y_values
        return df

    def compressed_plugin_data(
        self, table_data: List[Dict[str, int]]
    ) -> Union[EncodedFile, str]:
        return WebvizPluginABC.plugin_data_compress(
            [
                {
                    "filename": f"""{
                        self.component_unique_id(TableViewElement.Ids.TABLE).to_string()
                    }.csv""",
                    "content": self.download_data_df(table_data).to_csv(index=False),
                }
            ]
        )

    def set_callbacks(self) -> None:
        @callback(
            self.view_element_data_output(),
            self.view_element_data_requested(),
            State(
                self.component_unique_id(TableViewElement.Ids.TABLE).to_string(), "data"
            ),
        )
        def _download_data(
            data_requested: Union[int, None], table_data: List[Dict[str, int]]
        ) -> Union[EncodedFile, str]:
            if data_requested is None:
                raise PreventUpdate

            if not table_data:
                return "No data present in table"

            return self.compressed_plugin_data(table_data)


class TableViewSettingsGroup(SettingsGroupABC):
    class Ids(StrEnum):
        ORDER_SELECTOR = "order-selector"

    def __init__(self) -> None:
        super().__init__("Table orientation")

    def layout(self) -> List[Component]:
        return [
            wcc.RadioItems(
                id=self.register_component_unique_id(
                    TableViewSettingsGroup.Ids.ORDER_SELECTOR
                ),
                options=[
                    {
                        "label": "ASC",
                        "value": Order.ASC,
                    },
                    {
                        "label": "DESC",
                        "value": Order.DESC,
                    },
                ],
                value=Order.ASC,
            )
        ]


class TableView(ViewABC):
    class Ids(StrEnum):
        TEXT = "text"
        TABLE = "table"
        TABLE_SETTINGS = "table-settings"

    @dataclass
    class Slots:
        kindness_selector: Annotated[Input, str]
        power_selector: Annotated[Input, str]

    def __init__(self, data: List[Tuple[int, int]], slots: Slots) -> None:
        super().__init__("Table")
        self.data = data

        self.slots = slots

        self.table_view = TableViewElement(self.data)

        self.add_view_element(TextViewElement(), TableView.Ids.TEXT)
        self.add_view_element(self.table_view, TableView.Ids.TABLE)

        self.add_settings_group(
            TableViewSettingsGroup(), settings_group_id=TableView.Ids.TABLE_SETTINGS
        )

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.table_view.component_unique_id(
                    TableViewElement.Ids.TABLE
                ).to_string(),
                "data",
            ),
            self.slots.power_selector,
            Input(
                self.settings_group_unique_id(
                    TableView.Ids.TABLE_SETTINGS,
                    TableViewSettingsGroup.Ids.ORDER_SELECTOR,
                ),
                "value",
            ),
        )
        @callback_typecheck
        def _swap_order_and_power(power: int, order: Order) -> List[dict]:
            if power == 2:
                self.data = [(x, x * x) for x in range(0, 10)]
            else:
                self.data = [(x, x * x * x) for x in range(0, 10)]
            data = self.data.copy()
            if order == Order.DESC:
                data.reverse()
            return [{"x": d[0], "y": d[1]} for d in data]

        @callback(
            Output(
                self.view_element(TableView.Ids.TEXT)
                .component_unique_id(TextViewElement.Ids.TEXT)
                .to_string(),
                "children",
            ),
            self.slots.kindness_selector,
        )
        @callback_typecheck
        def _change_kindness(kindness: Kindness) -> Component:
            if kindness == Kindness.FRIENDLY:
                return [
                    html.H1("Hello"),
                    """
                    I am an example plugin.
                    Please have a look how views and settings are working in my environment =).
                    """,
                ]

            return [
                html.H1("Goodbye"),
                "I am a bloody example plugin. Leave me alone! =(",
            ]

        @callback(
            self.view_data_output(),
            self.view_data_requested(),
            State(
                self.table_view.component_unique_id(
                    TableViewElement.Ids.TABLE
                ).to_string(),
                "data",
            ),
        )
        def _download_data(
            data_requested: Union[int, None],
            table_data: List[Dict[str, int]],
        ) -> Union[EncodedFile, str]:
            if data_requested is None:
                raise PreventUpdate

            return self.table_view.compressed_plugin_data(table_data)



import math


def calculate_slider_step(
    min_value: float, max_value: float, steps: int = 100
) -> float:
    """Calculates a step value for use in e.g. dcc.RangeSlider() component
    that will always be rounded.

    The number of steps will be atleast the number
    of input steps, but might not be precisely the same due to use of the floor function.

    This function is necessary since there is currently no precision control in the underlying
    React component (https://github.com/react-component/slider/issues/275).

    """

    return 10 ** math.floor(math.log10((max_value - min_value) / steps))



import os
import sys
import warnings
from pathlib import Path
from typing import Dict, Set

import jinja2
import requests

from ..plugins import PLUGIN_PROJECT_METADATA
from ._pip_git_url import pip_git_url


PYPI_URL_ROOT = "https://pypi.org/"


def create_docker_setup(
    build_directory: Path, plugin_metadata: Dict[str, dict]
) -> None:
    """Creates a Docker setup in build_directory. The input dictionary plugin_metadata
    is a dictionary of plugin metadata only for the plugins used in the generated
    application. This is used in order to automatically include the necessary
    plugin projects as Python requirements in the generated Docker setup.
    """

    template_environment = jinja2.Environment(  # nosec
        loader=jinja2.PackageLoader("webviz_config", "templates"),
        undefined=jinja2.StrictUndefined,
        autoescape=False,
    )
    template = template_environment.get_template("Dockerfile.jinja2")

    distributions = {
        metadata["dist_name"]: PLUGIN_PROJECT_METADATA[metadata["dist_name"]]
        for metadata in plugin_metadata.values()
    }

    # Regardless of a standard webviz-config plugin is included in user's
    # configuration file, we still need to install the plugin framework webviz-config:
    distributions["webviz-config"] = PLUGIN_PROJECT_METADATA["webviz-config"]

    requirements = get_python_requirements(distributions)
    requirements.add("gunicorn")

    (build_directory / "requirements.txt").write_text(
        "\n".join(sorted(list(requirements)))
    )

    (build_directory / "Dockerfile").write_text(
        template.render(
            {
                "python_version_major": sys.version_info.major,
                "python_version_minor": sys.version_info.minor,
                "ssh_required": any("ssh://" in req for req in requirements),
            }
        )
    )


def get_python_requirements(distributions: dict) -> Set[str]:

    requirements = set()

    for dist_name, dist in distributions.items():

        requirements.update(
            [
                f"{dep}=={version}"
                for dep, version in PLUGIN_PROJECT_METADATA[dist_name][
                    "dependencies"
                ].items()
                if dep not in distributions
            ]
        )

        if dist["download_url"] is None and dist["source_url"] is None:
            warnings.warn(
                f"Plugin distribution {dist_name} has no download/source URL specified. "
                "Will therefore not automatically become part of built Docker image."
            )
            continue

        if dist["download_url"] is not None and dist["download_url"].startswith(
            PYPI_URL_ROOT
        ):
            pypi_data = requests.get(f"{PYPI_URL_ROOT}/pypi/{dist_name}/json").json()
            if dist["dist_version"] in pypi_data["releases"]:
                requirements.add(f"{dist_name}=={dist['dist_version']}")
                continue

            if dist["source_url"] is None:
                raise RuntimeError(
                    f"Could not find version {dist['dist_version']} of {dist_name} on PyPI. "
                    "Falling back to git source code is not possible since "
                    "project_urls['Source'] is not defined in setup.py."
                )

        requirements.add(
            pip_git_url(
                dist["dist_version"],
                source_url=os.environ.get(
                    "SOURCE_URL_" + dist_name.upper().replace("-", "_"),
                    dist["source_url"],
                ),
                git_pointer=os.environ.get(
                    "GIT_POINTER_" + dist_name.upper().replace("-", "_")
                ),
            )
        )

    return requirements



import copy
from enum import Enum
from pathlib import Path
from typing import Optional

from dash import Dash

from ._theme_class import WebvizConfigTheme


class WebvizRunMode(Enum):
    NON_PORTABLE = 1
    PORTABLE = 2
    BUILDING_PORTABLE = 3


class WebvizInstanceInfo:
    """Contains global information regarding the running webviz app instance, exposed
    globally through WEBVIZ_INSTANCE_INFO.

    Note that this class utilizes a two-stage initialization approach which renders it
    useless until the initialize() method has been called. It is assumed that
    initialization will be done early during application execution, typically as part
    of the webviz_app.py / jinja2 template.
    """

    __slots__ = (
        "_is_initialized",
        "_dash_app",
        "_run_mode",
        "_theme",
        "_storage_folder",
    )

    def __init__(self) -> None:
        self._is_initialized: bool = False
        self._dash_app: Optional[Dash] = None
        self._run_mode: Optional[WebvizRunMode] = None
        self._theme: Optional[WebvizConfigTheme] = None
        self._storage_folder: Optional[Path] = None

    def initialize(
        self,
        dash_app: Dash,
        run_mode: WebvizRunMode,
        theme: WebvizConfigTheme,
        storage_folder: Path,
    ) -> None:
        """This function is responsible for the actual initialization of the object instance.
        None of the access methods in this class are valid until initialize() has been called.
        This function will be called as part of the webviz_app.py / jinja2 template.
        """
        if self._is_initialized:
            raise RuntimeError("Registry already initialized")

        if not isinstance(dash_app, Dash):
            raise TypeError("dash_app must be of type Dash")
        self._dash_app = dash_app

        if not isinstance(run_mode, WebvizRunMode):
            raise TypeError("run_mode must be of type WebvizRunMode")
        self._run_mode = run_mode

        if not isinstance(theme, WebvizConfigTheme):
            raise TypeError("theme must be of type WebvizConfigTheme")
        self._theme = theme

        if not isinstance(storage_folder, Path):
            raise TypeError("storage_folder must be of type Path")
        self._storage_folder = storage_folder

        self._is_initialized = True

    @property
    def dash_app(self) -> Dash:
        if not self._is_initialized or self._dash_app is None:
            raise RuntimeError("WebvizInstanceInfo is not yet initialized")

        return self._dash_app

    @property
    def run_mode(self) -> WebvizRunMode:
        if not self._is_initialized or self._run_mode is None:
            raise RuntimeError("WebvizInstanceInfo is not yet initialized")

        return self._run_mode

    @property
    def theme(self) -> WebvizConfigTheme:
        if not self._is_initialized or self._theme is None:
            raise RuntimeError("WebvizInstanceInfo is not yet initialized")

        # The theme class is mutable, so return a copy
        return copy.deepcopy(self._theme)

    @property
    def storage_folder(self) -> Path:
        if not self._is_initialized or self._storage_folder is None:
            raise RuntimeError("WebvizInstanceInfo is not yet initialized")

        return self._storage_folder


WEBVIZ_INSTANCE_INFO = WebvizInstanceInfo()



import sys
from pathlib import Path


def user_data_dir() -> Path:
    """Returns platform specific path to store user application data"""

    if sys.platform == "win32":
        return Path.home() / "AppData" / "webviz"

    if sys.platform == "darwin":
        return Path.home() / "Library" / "Application Support" / "webviz"

    return Path.home() / ".local" / "share" / "webviz"
